% Code for Machine Learning Lecture 7
%
% Based on the EM Code by Thomas Schn written on Feb. 21, 2011
% by Umut Orguner on March 02, 2011

clear;
close all;

myeps=1e-6;%a very small number
fontsize=15;

% Gaussian mixture with 3 terms
pi1    = 0.3;
mu1    = [4 4.5];
Sigma1 = [1.2 0.6; 0.6 0.5];

pi2    = 0.5;
mu2    = [8 1];  
Sigma2 = [1 0; 0 1];

pi3    = 0.2;
mu3    = [9 8];
Sigma3 = [0.6 0.5; 0.5 1.5];

%mesh-plot the GM
[X1,X2] = meshgrid(linspace(0,14,200)', linspace(-4,12,200)');
X = [X1(:) X2(:)];
p = pi1*mvnpdf(X, mu1, Sigma1) + pi2*mvnpdf(X, mu2, Sigma2) + pi3*mvnpdf(X, mu3, Sigma3);
figure(1)
surf(X1,X2,reshape(p,200,200));
xlabel('x_1')
ylabel('x_2')

%contour plot the GM
figure(2)
contour(X1,X2,reshape(p,200,200),100);
xlabel('x_1')
ylabel('x_2')

% Generate N samples from the above Gaussian mixture
N = 1000;            % Number of samples
x = zeros(2,N);
piemp = zeros(3,1);  % Empirical estimate of pi
for i=1:N
  u = rand(1);
  if u<pi1
    piemp(1) = piemp(1) + 1;
    n = mu1' + chol(Sigma1)*randn(2,1);
  elseif and(u>=pi1, u<pi1+pi2)
    piemp(2) = piemp(2) + 1;
    n = mu2' + chol(Sigma2)*randn(2,1);
  else
    piemp(3) = piemp(3) + 1;
    n = mu3' + chol(Sigma3)*randn(2,1);
  end;
  x(:,i) = n;
end;
piemp = piemp/N;
%end of sample generation

%plot the samples
figure(3)
subplot(1,2,1)
plot(x(1,:),x(2,:),'o','MarkerSize',1.5,'MarkerEdgeColor','k','MarkerFaceColor','k');
axis([0 14 -4 12])
xlabel('x_1')
ylabel('x_2')
hold on
subplot(1,2,2)
plot(x(1,:),x(2,:),'o','MarkerSize',1.5,'MarkerEdgeColor','k','MarkerFaceColor','k');
axis([0 14 -4 12])
xlabel('x_1')
ylabel('x_2')
hold on

% EM for estimating Gaussian mixtures
maxIter = 400;     % Maximum number of iterations
K=10; %number of components

%EM Initialize
muEM=mvnrnd([8 5],25*eye(2),K)';%random means
sigmaEM=zeros(2,2,K);%covariances
sigmaEM(1,1,:)=1;%set covariances to identity matrix
sigmaEM(2,2,:)=1;%set covariances to identity matrix
piEM=ones(1,K)/K;%equal probabilities

%VB Initialize
VB_alpha0=1e-10;
VB_m0=zeros(2,1);
VB_W0=10^5*eye(2);
VB_v0=0.01;
VB_beta0=0.01;



muVB=muEM;%random means
sigmaVB=sigmaEM;%initial covariances
piVB=piEM;%equal probabilities for EM and VB

%some VB parameters
pitildeVB=ones(1,K)/K;
lambdatildeVB=ones(1,K);
vkVB=VB_v0*ones(1,K);
mkVB=muVB;
betakVB=VB_beta0*ones(1,K);
WkVB=sigmaVB;
alphakVB=VB_alpha0*ones(1,K);


rEM=zeros(K,N);%EM responsibilities
rVB=zeros(K,N);%VB responsibilities
hEM=[];%EM plot handle
hVB=[];%EM plot handle
for i = 1:maxIter
  disp(['Iteration: ', num2str(i)]);
 
  if (i==1)
        %Plot Estimates
  figure(3)
  delete(hEM)
  subplot(1,2,1);
  title(['EM Iteration: ', num2str(i), '  No of Components Left: ',num2str(sum(piEM>1e-5))],'FontSize',fontsize);   
  hEM=confellipse2(muEM(:,piEM>1e-5),sigmaEM(:,:,piEM>1e-5),piEM(piEM>1e-5),'b',gca,0.95,2);
    
  delete(hVB)
  subplot(1,2,2)
  title(['VB Iteration: ', num2str(i), '  No of Components Left: ',num2str(sum(piVB>1e-5))],'FontSize',fontsize);  
  hVB=confellipse2(muVB(:,piVB>1e-5),sigmaVB(:,:,piVB>1e-5),piVB(piVB>1e-5),'b',gca,0.95,2);
  pause%wait
  end
  
  %EM Implementation
  %E-step 
  for k=1:K,
      rEM(k,:)=piEM(k)*mvnpdf(x',muEM(:,k)',sigmaEM(:,:,k)+myeps*eye(2))';
  end
  denominator=sum(rEM,1);
  %normalize
  for k=1:K,
      rEM(k,:)=rEM(k,:)./denominator;      
  end
  %M-step
  NkEM= sum(rEM,2);
  piEM= NkEM/N';
  for k = 1:K,
    muEM(:,k)=x*rEM(k,:)'/NkEM(k);
    sigmaEM(:,:,k)=(x.*repmat(rEM(k,:),2,1)*x')/NkEM(k)-muEM(:,k)*muEM(:,k)'; 
    sigmaEM(:,:,k)=0.5*(sigmaEM(:,:,k)+sigmaEM(:,:,k)');
  end;
  %End of EM
  
  %VB implementation based on Bishop
  for k=1:K,
      sqrtWkVB=cholcov(WkVB(:,:,k)+myeps*eye(2));
      rVB(k,:)=pitildeVB(k)*sqrt(lambdatildeVB(k))*exp(-2/2/betakVB(k)-0.5*vkVB(k)*sum((sqrtWkVB*(x-repmat(mkVB(:,k),1,N))).^2,1));
  end
  denominator=sum(rVB,1);
  %normalize
  for k=1:K,
      rVB(k,:)=rVB(k,:)./denominator;      
  end
  %M-step
  NkVB= sum(rVB,2);
  for k = 1:K,
    if NkVB(k)~=0,  
        muVB(:,k)=x*rVB(k,:)'/NkVB(k);
        sigmaVB(:,:,k)=(x.*repmat(rVB(k,:),2,1)*x')/NkVB(k)-muVB(:,k)*muVB(:,k)'; 
        sigmaVB(:,:,k)=0.5*(sigmaVB(:,:,k)+sigmaVB(:,:,k)');
    end
  end;
  betakVB=VB_beta0+NkVB';
  vkVB=VB_v0+NkVB+1;
  alphakVB=VB_alpha0+NkVB;
  for k=1:K,
      mkVB(:,k)=(VB_beta0*VB_m0+NkVB(k)*muVB(:,k))./betakVB(k);
      WkVB(:,:,k)=inv(inv(VB_W0)+NkVB(k)*sigmaVB(:,:,k)+VB_beta0*NkVB(k)*(muVB(:,k)-VB_m0)*(muVB(:,k)-VB_m0)'/(VB_beta0+NkVB(k)));
      WkVB(:,:,k)=0.5*(WkVB(:,:,k)+WkVB(:,:,k)');
      lambdatildeVB(k)=exp(sum(psi((vkVB(k)+1-(1:2))/2))+2*log(2)+log(det(WkVB(:,:,k))));      
  end    
  pitildeVB=exp(psi(alphakVB)-psi(sum(alphakVB)));  
  piVB=NkVB/N';  
  %End of VB
  
  %Plot Estimates
  figure(3)
  delete(hEM)
  subplot(1,2,1);
  title(['EM Iteration: ', num2str(i), '  No of Components Left: ',num2str(sum(piEM>1e-5))],'FontSize',fontsize);  
  hEM=confellipse2(muEM(:,piEM>1e-5),sigmaEM(:,:,piEM>1e-5),piEM(piEM>1e-5),'b',gca,0.95,2);%show EM components with only sufficiently high responsibilities.
    
  delete(hVB)
  subplot(1,2,2)
  title(['VB Iteration: ', num2str(i), '  No of Components Left: ',num2str(sum(piVB>1e-5))],'FontSize',fontsize);  
  hVB=confellipse2(muVB(:,piVB>1e-5),sigmaVB(:,:,piVB>1e-5),piVB(piVB>1e-5),'b',gca,0.95,2);%show VB components with only sufficiently high responsibilities.
  
  pause(0.2)
end;
hold off






