function varargout=confellipse2(x,P,weight,col,ax,Pg,linewidth)

%CONFELLIPSE2 Adds a confidence ellipse of level (%) to a SIG plot
% confellipse2(x,P,col,ax,level,linewidth)
% x is (N,2), Px is (N,2,2)


%   Fredrik Gustafsson 08-May-2006
%   Copyright (c) 1994-2006 by COMSOL AB
%   $Revision: 1.2 $ $Date: 2007/11/29 16:06:31 $
if nargin<6, linewidth=1; end
if nargin<5, Pg=99; end
[nx,N]=size(x);

if nx~=2
    error('nx must be 2');
end
phi=linspace(0,2*pi,100);
z=[cos(phi);sin(phi)];
if Pg<1
    c=sqrt(chi2inv(Pg,2));
else
    error('Gate probability Pg must be smaller than 1')
end

h=zeros(1,N);
htext=zeros(1,N);
if (N==1)
    [U,S,V]=svd(P);
    Sroot=sqrt(S);
    Ph=U*Sroot;
    h=plot(x(1)+c*Ph(1,:)*z,x(2)+c*Ph(2,:)*z,[col,'-'],'parent',ax,'linewidth',linewidth);
else
    for k=1:N
        [U,S,V]=svd(squeeze(P(:,:,k)));
        Sroot=sqrt(S);
        Ph=U*Sroot;
        h(k)=plot(x(1,k)+c*Ph(1,:)*z,x(2,k)+c*Ph(2,:)*z,[col,'-'],'parent',ax,'LineWidth',linewidth);
        htext(k)=text(x(1,k)+c*Ph(1,1),x(2,k)+c*Ph(2,1),num2str(weight(k)),'parent',ax,'LineWidth',linewidth,'FontSize',12,'FontWeight','bold','Color','r');
    end
end
if (nargout==1)
    varargout={[h htext]};
end
